var App        = {};
App.Router     = {};
App.View       = {};
App.Model      = {};
App.Collection = {};
App.Util       = {};

App.Util.Board = (function () {
  function Board(id) {
    this.id       = id;
    this.canvas   = document.getElementById(this.id);
    this.context  = this.canvas.getContext('2d');
    this.width    = this.canvas.width;
    this.height   = this.canvas.height;
    this.degrees  = Math.PI / 10.0;
    this.add      = Math.PI / 20.0;
    this.ring = {
      count  : 7,
      radius : (this.width > this.height) ? this.height / 14.5 : this.width / 14.5
    };
    this.center  = {
      x : this.width / 2,
      y : this.height / 2
    };
    this.color = {
      black  : '#222222',
      red    : '#CC3300',
      green  : '#5cb85c',
      white  : '#ffffff',
      gray   : '#080808',
      hover  : '#428bca',
      active : '#5bc0de'
    };
  }
  Board.prototype.reset = function () {
    this.context.beginPath();
    this.context.fillStyle   = 'rgba(0,0,0,0)';
    this.context.strokeStyle = 'rgba(0,0,0,0)';
    this.context.shadowColor = 'rgba(0,0,0,0)';
    this.context.shadowBlur  = 0;
  };
  Board.prototype.clear = function () {
    this.reset();
    this.context.clearRect(0, 0, this.width, this.height);
  };
  Board.prototype.arc = function (r, sa, sb, color) {
    var s, x, y;
    this.reset();
    this.context.moveTo(this.center.x, this.center.y);
    for (s = sa; s < sb; s += 1 / r) {
      x = r * Math.cos(s) + this.center.x;
      y = r * Math.sin(s) + this.center.y;
      this.context.lineTo(x, y);
    }
    this.context.moveTo(this.center.x, this.center.y);
    this.context.closePath();
    this.context.fillStyle = color;
    this.context.fill();
  };
  Board.prototype.spider = function () {
    var s, x, y;
    this.reset();
    this.context.moveTo(this.center.x, this.center.y);
    for (s = 0; s < Math.PI * 2; s += this.degrees) {
      x = this.ring.radius * this.ring.count * Math.cos(s + this.add) + this.center.x;
      y = this.ring.radius * this.ring.count * Math.sin(s + this.add) + this.center.y;
      this.context.moveTo(this.center.x, this.center.y);
      this.context.lineTo(x, y);
    }
    this.context.strokeStyle = this.color.gray;
    this.context.stroke();
  };
  Board.prototype.circle = function (ring, color) {
    var i, r, s, x, y;
    for (i = this.ring.count - 1; i >= 0; i -= 1) {
      r = this.ring.radius * (1 + i);
      this.reset();
      for (s = 0.0; s < 2 * Math.PI; s += 1 / r) {
        x = r * Math.cos(s) + this.center.x;
        y = r * Math.sin(s) + this.center.y;
        if (s === 0) {
          this.context.moveTo(x, y);
        } else {
          this.context.lineTo(x, y);
        }
      }
      this.context.closePath();
      if (i === 0) {
        this.context.fillStyle = (i === ring) ? color : this.color.black;
        this.context.fill();
      } else if (i === 1) {
        this.context.fillStyle = (i === ring) ? color : this.color.red;
        this.context.fill();
      }
      this.context.strokeStyle = this.color.gray;
      this.context.stroke();
    }
  };
  Board.prototype.text = function () {
    var i, s, x, y,
      scoreText = [6, 10, 15, 2, 17, 3, 19, 7, 16, 8, 11, 14, 9, 12, 5, 20, 1, 18, 4, 13];
    this.reset();
    this.context.moveTo(this.center.x, this.center.y);
    this.context.fillStyle    = "white";
    this.context.textBaseline = "middle";
    this.context.textAlign    = "center";
    this.context.font         = this.ring.radius - 16 + "px sans-serif";
    for (i = 0, s = 0; s < Math.PI * 2; s += this.degrees, i += 1) {
      x = this.ring.radius * (this.ring.count - 0.5) * Math.cos(s) + this.center.x;
      y = this.ring.radius * (this.ring.count - 0.5) * Math.sin(s) + this.center.y;
      this.context.fillText(scoreText[i], x, y);
    }
  };
  Board.prototype.draw = function () {
    var i,
        red     = this.color.red,
        green   = this.color.green,
        black   = this.color.black,
        white   = this.color.white,
        radius  = this.ring.radius,
        count   = this.ring.count,
        degrees = this.degrees,
        add     = this.add;
    this.clear();
    for (i = 0; i < 20; i += 1) {
      this.arc(radius * count, i * degrees + add, (i + 1) * degrees + add, black);
      if (i % 2 === 0) {
        this.arc(radius * (count - 1), i * degrees + add, (i + 1) * degrees + add, red);
        this.arc(radius * (count - 2), i * degrees + add, (i + 1) * degrees + add, black);
        this.arc(radius * (count - 3), i * degrees + add, (i + 1) * degrees + add, red);
        this.arc(radius * (count - 4), i * degrees + add, (i + 1) * degrees + add, black);
      } else {
        this.arc(radius * (count - 1), i * degrees + add, (i + 1) * degrees + add, green);
        this.arc(radius * (count - 2), i * degrees + add, (i + 1) * degrees + add, white);
        this.arc(radius * (count - 3), i * degrees + add, (i + 1) * degrees + add, green);
        this.arc(radius * (count - 4), i * degrees + add, (i + 1) * degrees + add, white);
      }
    }
    this.spider();
    this.circle();
    this.text();
  };
  Board.prototype.result = function (e) {
    // @TODO why can't use canvas.offset...
    // var x      = e.pageX - this.canvas.offsetLeft - this.center.x,
    //     y      = e.pageY - this.canvas.offsetTop  - this.center.y,

    // HiDPI support
    var x      = e.pageX - $('#' + this.id).offset().left - (this.center.x/2),
        y      = e.pageY - $('#' + this.id).offset().top  - (this.center.y/2),
        norm   = Math.sqrt(x * x + y * y),
        ux     = x / norm,
        uy     = y / norm,
        s      = Math.atan2(-ux, uy) + Math.PI / 2 - this.add,
        spider = Math.floor(s / this.degrees),
        ring   = Math.floor(norm / (this.ring.radius/2)), // HiDPI support
        points = [10, 15, 2, 17, 3, 19, 7, 16, 8, 11, 14, 9, 12, 5, 20, 1, 18, 4, 13, 6],
        point  = 0,
        label  = '';
    if (0 > spider) {spider += 20; }
    if (ring === 0) {
      point = 50;
      label = 'IN BULL';
      position = 'bull';
    } else if (ring === 1) {
      point = 25;
      label = 'OUT BULL';
      position = 'bull';
    } else if (ring === 3) {
      point = points[spider] * 3;
      label = 'TRIPLE ' + points[spider];
      position = 'triple';
    } else if (ring === 5) {
      point = points[spider] * 2;
      label = 'DOUBLE ' + points[spider];
      position = 'double';
    } else if (ring === 2 || ring === 4) {
      point = points[spider];
      label = 'SINGLE ' + points[spider];
      position = 'single';
    } else if (ring > 6) {
      point = 0;
      label = 'OUT';
      position = 'out';
    } else if (ring === 6) {
      return false;
    }
    return {
      label   : label,
      point   : point,
      position: position,
      area    : points[spider],
      ring    : ring,
      spider  : spider,
    };
  };
  Board.prototype.overwrite = function (e, color) {
    var i, temporary_color,
        result  = this.result(e),
        ring    = result.ring,
        spider  = result.spider,
        red     = this.color.red,
        black   = this.color.black,
        green   = this.color.green,
        white   = this.color.white,
        degrees = this.degrees,
        count   = this.ring.count,
        radius  = this.ring.radius,
        add     = this.add;
    this.clear();
    if (result.ring > 6) {
      this.context.fillStyle = color;
      this.context.fillRect(0, 0, this.width, this.height);
      this.reset();
    }
    for (i = 0; i < 20; i += 1) {
      this.arc(radius * count, i * degrees + add, (i + 1) * degrees + add, black);
      if (i % 2 === 0) {
        temporary_color = (ring + 1 === count - 1 && spider === i) ? color : red;
        this.arc(radius * (count - 1), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 2 && spider === i) ? color : black;
        this.arc(radius * (count - 2), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 3 && spider === i) ? color : red;
        this.arc(radius * (count - 3), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 4 && spider === i) ? color : black;
        this.arc(radius * (count - 4), i * degrees + add, (i + 1) * degrees + add, temporary_color);
      } else {
        temporary_color = (ring + 1 === count - 1 && spider === i) ? color : green;
        this.arc(radius * (count - 1), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 2 && spider === i) ? color : white;
        this.arc(radius * (count - 2), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 3 && spider === i) ? color : green;
        this.arc(radius * (count - 3), i * degrees + add, (i + 1) * degrees + add, temporary_color);
        temporary_color = (ring + 1 === count - 4 && spider === i) ? color : white;
        this.arc(radius * (count - 4), i * degrees + add, (i + 1) * degrees + add, temporary_color);
      }
    }
    this.spider();
    this.circle(ring, color);
    this.text();
  };
  return Board;
}());

App.Model.Player = Backbone.Model.extend({
  initialize: function () {
    this.name = 'Guest';
    this.score = {
      count_up: [],
      501: [],
      cricket: []
    };
  },
  rating: function (game) {
    return 'Null';
  },
  average: function (game) {
    return 'Null';
  }
});

App.Model.Score = Backbone.Model.extend({
  initialize: function () {
    this.player_cid    = '';
    this.label = '';
    this.point = 0;
    this.position = '';
  }
});

var CountUp = Backbone.Collection.extend({
  model: App.Model.Score,
  initialize: function () {
    this.on('add', function (score) {
      if (this.length > App.Collection.Players.length * 8 * 3) {
        this.pop(score);
        if (window.confirm('GAME END')) {
          App.Router.CountUp.End();
        }
      }
    });
  },
  now_player: function (options) {
    options = options || {};
    var count = _.isNumber(options.count) ? options.count : this.length;
    count = (0 >= count) ? 1 : count;
    var players_length = App.Collection.Players.length;
    while (count >= players_length * 3) {
      count -= players_length * 3;
    }
    var index = Math.floor(count / 3);
    return App.Collection.Players.models[index];
  },
  now_round: function (options) {
    options = options || {};
    var count = _.isNumber(options.count) ? options.count : this.length;
    count = (0 >= count) ? 1 : count;
    var players_length = App.Collection.Players.length;
    var round = Math.floor(count / players_length / 3);
    return round;
  },
  game_data: function () {
    var count   = this.length;
    var players = App.Collection.Players.models;
    var throw_count = 0;
    var game_data = [];
    var now_player, now_round, player_index = 0;
    for (var i = 0; count > i; i++, throw_count++) {
      if (throw_count === 3) throw_count = 0;
      now_round = this.now_round({count: i});
      now_player = players.indexOf(this.now_player({count: i}));
      game_data[now_player] = game_data[now_player] || [];
      game_data[now_player][now_round] = game_data[now_player][now_round] || [];
      game_data[now_player][now_round][throw_count] = game_data[now_player][now_round][throw_count] || [];
      game_data[now_player][now_round][throw_count] = this.models[i];
    }
    return game_data;
  },
  game_score: function () {
    var players = App.Collection.Players.models;
    var game_data = this.game_data();
    var now_player = players.indexOf(this.now_player({count: this.length - 1}));
    var game_score = {};
    var i = 0;
    game_score.now_player = now_player;
    game_score.dart_result = [];
    game_score.score_round = [];
    game_score.player_score = [];
    // Dart Result
    for (i = 0; 3 > i; i++) {
      try {
        game_score.dart_result[i] = _.last(game_data[now_player])[i].get('label');
      } catch(e) {
        game_score.dart_result[i] = '- ' + (i + 1) +' -';
      }
    }
    // Score Round
    for (i = 0; 8 > i; i++) {
      try {
          game_score.score_round[i]  = game_data[now_player][i][0].get('point') || 0;
          game_score.score_round[i] += game_data[now_player][i][1].get('point') || 0;
          game_score.score_round[i] += game_data[now_player][i][2].get('point') || 0;
      } catch(e) {
        game_score.score_round[i] = _.isNumber(game_score.score_round[i]) ?
          game_score.score_round[i] : '';
      }
    }
    // Player Score
    for (var index = 0; players.length > index; index++) {
      for (i = 0; 8 > i; i++) {
        try {
            game_score.player_score[index] = game_score.player_score[index] || 0;
            game_score.player_score[index] += game_data[index][i][0].get('point') || 0;
            game_score.player_score[index] += game_data[index][i][1].get('point') || 0;
            game_score.player_score[index] += game_data[index][i][2].get('point') || 0;
        } catch(e) {
          game_score.player_score[index] = _.isNumber(game_score.player_score[index]) ?
            game_score.player_score[index] : 0;
        }
      }
    }
    return game_score;
  }
});

App.Collection.CountUp = new CountUp();

var Players = Backbone.Collection.extend({
  model: App.Model.Player,
  initialize: function () {
    this.on('add', function (player) {
      if (this.length > 4) {
        this.pop(player);
      }
    });
  }
});

App.Collection.Players = new Players();

App.View.About = Backbone.View.extend({
  el: "body",
  template: JST['about'],
  events: {
  },
  initialize: function() {
    _.bindAll(this, 'render', 'after_render');
    var that = this;
    this.render = _.wrap(this.render, function(render) {
      render();
      that.after_render();
      return that;
    });
  },
  render: function() {
    var html = this.template();
    this.$el.html(html);
    $(window).off('resize').on('resize', _.bind(this.after_render, this));
    return this;
  },
  after_render: function (){
    var width = $(window).width();
    var height = Math.floor(width * 1242 / 2038);
    var top = 0;
    if (height > 230) {
      top = height / 2 * -1;
    }
    $('#gear').css('margin-top', top);
  }
});

App.View.CountUp = Backbone.View.extend({
  el: "body",
  template: JST['count_up'],
  events: {
    'mousemove #board': function (e) {
      this.board.overwrite(e, this.board.color.hover);
    },
    'mousedown #board': function (e) {
      this.board.overwrite(e, this.board.color.active);
    },
    'mouseup #board': function (e) {
      this.board.overwrite(e, this.board.color.hover);
    },
    'mouseout #board': function () {
      this.board.draw();
    },
    'click #undo': function () {
      App.Collection.CountUp.pop();
      this.render();
    },
    'click #board': function (e) {
      var result = this.board.result(e);
      if (result === false) { return; }
      var score = {
        point: result.point,
        position: result.position,
        label: result.label
      };
      this.add_score(score);
      this.render();
      this.board.overwrite(e, this.board.color.hover);
    }
  },
  initialize: function() {
    _.bindAll(this, 'render', 'after_render');
    var that = this;
    this.render = _.wrap(this.render, function(render) {
      render();
      that.after_render();
      return that;
    });
  },
  render: function() {
    var html = this.template({
      players: App.Collection.Players,
      game_score: App.Collection.CountUp.game_score()
    });
    this.$el.html(html);
    $(window).off('resize').on('resize', _.bind(this.after_render, this));
    return this;
  },
  after_render: function() {
    var side = {
      width:  this.$('.center_column').width(),
      height: this.$(".left_column").height() - 20
    };
    this.$('#board').width(side.width + 'px');
    this.$('#board').height(side.height + 'px');

    // HiDPI support
    this.$('#board').attr({width:side.width * 2 + 'px'});
    this.$('#board').attr({height:side.height * 2 + 'px'});
    this.board = new App.Util.Board('board');
    this.board.draw();
  },
  add_score: function(obj) {
    var now_player   = App.Collection.CountUp.now_player();
    obj.player_cid = now_player.cid;

    var score = new App.Model.Score(obj);
    App.Collection.CountUp.add(score);
  },
});

App.View.GameEnd = Backbone.View.extend({
  el: "body",
  template: JST['game_end'],
  initialize: function() {
    this.game = 'count_up';
  },
  render: function() {
    var html = this.template({
      game: this.game,
      players: App.Collection.Players
    });
    this.$el.html(html);
    return this;
  }
});

App.View.GameMenu = Backbone.View.extend({
  el: "body",
  template: JST['game_menu'],
  events: {
    'click #add_player'   : 'add_player',
    'click .remove_player': 'remove_player',
    'change #game'        : 'change_game',
    'click #game_start'   : 'game_start'
  },
  initialize: function() {
    this.game = 'count_up';
  },
  game_start: function() {
    App.Collection.CountUp.reset();
  },
  change_game: function (){
    this.game = this.$('#game').val();
    this.render();
  },
  add_player: function () {
    var player = new App.Model.Player();
    App.Collection.Players.add(player);
    this.render();
  },
  remove_player: function (e) {
    var self = $(e.toElement);
    var index = this.$('.remove_player').index(self);
    var player = App.Collection.Players.models[index];
    App.Collection.Players.remove(player);
    this.render();
  },
  render: function() {
    var html = this.template({
      game: this.game,
      players: App.Collection.Players
    });
    this.$el.html(html);
    return this;
  }
});

App.View.Home = Backbone.View.extend({
  el: "body",
  template: JST['home'],
  events: {
  },
  initialize: function() {
  },
  render: function() {
    var html = this.template();
    this.$el.html(html);
    return this;
  }
});

var About = Backbone.Router.extend({
  initialize: function () {
    this.AboutView = new App.View.About();
  },
  routes: {
    'about' : 'about',
  },
  about: function () {
    this.AboutView.render();
  },
});

App.Router.About = new About();

var CountUp = Backbone.Router.extend({
  initialize: function () {
    this.CountUpView  = new App.View.CountUp();
  },
  routes: {
    'count_up'  : 'Start',
  },
  Start: function () {
    if (App.Collection.Players.length > 0) {
      this.CountUpView.render();
    } else {
      this.navigate('game_menu', {trigger: true, replace: true});
    }
  },
  End: function () {
    this.navigate('game_end', {trigger: true, replace: true});
  }
});
App.Router.CountUp = new CountUp();

var GameEnd = Backbone.Router.extend({
  initialize: function () {
    this.GameEndView = new App.View.GameEnd();
  },
  routes: {
    'game_end': 'game_end',
  },
  game_end: function () {
    this.GameEndView.render();
  }
});

App.Router.GameEnd = new GameEnd();

var GameMenu = Backbone.Router.extend({
  initialize: function () {
    this.GameMenuView  = new App.View.GameMenu();
  },
  routes: {
    'game_menu'  : 'game_menu',
  },
  game_menu: function () {
    this.GameMenuView.render();
  }
});
App.Router.GameMenu = new GameMenu();

var Home = Backbone.Router.extend({
  initialize: function () {
    this.HomeView  = new App.View.Home();
  },
  routes: {
    ''      : 'home',
    'home'  : 'home',
  },
  home: function () {
    this.HomeView.render();
  }
});

App.Router.Home = new Home();

Backbone.history.start({});
