App.View.CountUp = Backbone.View.extend({
  el: "body",
  template: JST['count_up'],
  events: {
    'mousemove #board': function (e) {
      this.board.overwrite(e, this.board.color.hover);
    },
    'mousedown #board': function (e) {
      this.board.overwrite(e, this.board.color.active);
    },
    'mouseup #board': function (e) {
      this.board.overwrite(e, this.board.color.hover);
    },
    'mouseout #board': function () {
      this.board.draw();
    },
    'click #undo': function () {
      App.Collection.CountUp.pop();
      this.render();
    },
    'click #board': function (e) {
      var result = this.board.result(e);
      if (result === false) { return; }
      var score = {
        point: result.point,
        position: result.position,
        label: result.label
      };
      this.add_score(score);
      this.render();
      this.board.overwrite(e, this.board.color.hover);
    }
  },
  initialize: function() {
    _.bindAll(this, 'render', 'after_render');
    var that = this;
    this.render = _.wrap(this.render, function(render) {
      render();
      that.after_render();
      return that;
    });
  },
  render: function() {
    var html = this.template({
      players: App.Collection.Players,
      game_score: App.Collection.CountUp.game_score()
    });
    this.$el.html(html);
    $(window).off('resize').on('resize', _.bind(this.after_render, this));
    return this;
  },
  after_render: function() {
    var side = {
      width:  this.$('.center_column').width(),
      height: this.$(".left_column").height() - 20
    };
    this.$('#board').width(side.width + 'px');
    this.$('#board').height(side.height + 'px');

    // HiDPI support
    this.$('#board').attr({width:side.width * 2 + 'px'});
    this.$('#board').attr({height:side.height * 2 + 'px'});
    this.board = new App.Util.Board('board');
    this.board.draw();
  },
  add_score: function(obj) {
    var now_player   = App.Collection.CountUp.now_player();
    obj.player_cid = now_player.cid;

    var score = new App.Model.Score(obj);
    App.Collection.CountUp.add(score);
  },
});
